unit Spline2D;
{ Example on how to define a new primitive. }

interface

uses
 Graphics, Classes, Prim2D, Obj2D, GrLib2D;

type
  TBSpline2D = class(TPolyline2D)
  protected
    FPointsInCurve: Word;
    FOrder: Byte;
    procedure GetCurvePoints(var RV: TVect2D); virtual;
  public
    constructor Create(ID: LongInt; const Pts: array of TPoint2D);
    constructor CreateFromStream(ID: Byte; Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
    procedure Draw(const VT: TTransf2D; const Cnv: TCanvas); override;
    function OnMe(Pt: TPoint2D; Aperture: TRealType;
                  var Distance: TRealType): Integer; override;
    property PointsInCurve: Word read FPointsInCurve write FPointsInCurve;
    property Order: Byte read FOrder write FOrder;
  end;

implementation

{ The below function is copied by a book on Graphics Programming
  from Mc GrawHill. }

function Knot(I, K, N: Integer): Integer;
begin
  if I < K then
   Result := 0
  else if I > N then
   Result := N - K + 2
  else
   Result := I - K + 1;
end;

function NBlend(I, K, OK, N: Integer; U: TRealType): TRealType;
var
  T: Integer;
  V: TRealType;
begin
  if K = 1 then
   begin
     V := 0;
     if (Knot(i, OK, N) <= u) and (u < Knot(i + 1, OK, N)) then
      V := 1;
   end
  else
   begin
     V := 0;
     T := Knot(I + K - 1, OK, N) - Knot(I, OK, N);
     if T <> 0 then
      V := (U - Knot(I, OK, N)) * NBlend(I, K -1, OK, N, U) / T;
     T := Knot(I + K, OK, N) - Knot(I + 1, OK, N);
     if T <> 0 then
      V := V + (Knot(I + K, OK, N) - U) * NBlend(I + 1, K - 1, OK, N, U) / T;
   end;
  Result := V;
end;

function BSpline2D(U: TRealType; N, K: Integer; const Points: TVect2D): TPoint2D;
var
  I: Integer;
  B: TRealType;
begin
  Result := Point2D(0.0, 0.0);
  for I := 0 to N do
   begin
     B := NBlend(I, K, K, N, U);
     Result.X := Result.X + Points[I].X * B;
     Result.Y := Result.Y + Points[I].Y * B;
   end;
end;

{ It doesn't create RV. }
procedure TBSpline2D.GetCurvePoints(var RV: TVect2D);
var
  Cont: Integer;
  Delta, CurrAngle, CX, RX, CY, RY: TRealType;
begin
  for Cont := 0 to FPointsInCurve - 1 do
   RV.Add(BSpline2D(Cont / FPointsInCurve * (FPoints.Count - 2),
                    FPoints.Count - 1, FOrder, FPoints));
  if FPoints.Count > FOrder then
   { Add a point to close the spline if the number of point is greater than
     the order of the spline. }
   RV.Add(FPoints[FPoints.Count - 1]);
end;

constructor TBSpline2D.Create(ID: LongInt; const Pts: array of TPoint2D);
begin
  inherited Create(ID, Pts);
  FPointsInCurve := 100;
  FOrder := 3; { The default spline is cubic. }
end;

constructor TBSpline2D.CreateFromStream(ID: Byte; Stream: TStream);
begin
  { Load the standard properties }
  inherited CreateFromStream(ID, Stream);
  with Stream do
   begin
     { Load the particular properties. }
     Read(FOrder, SizeOf(FOrder));
     Read(FPointsInCurve, SizeOf(FPointsInCurve));
   end;
end;

procedure TBSpline2D.SaveToStream(Stream: TStream);
begin
  { Save the standard properties }
  inherited SaveToStream(Stream);
  with Stream do
   begin
     { Save the particular properties. }
     Write(FOrder, SizeOf(FOrder));
     Write(FPointsInCurve, SizeOf(FPointsInCurve));
   end;
end;

procedure TBSpline2D.Draw(const VT: TTransf2D; const Cnv: TCanvas);
var
  TmpVect: TVect2D;
  TmpFlag: Boolean;
  TmpPt: TPoint2D;
  TmpTransf: TTransf2D;
  Cont: Integer;
begin
  { I've decided to make the spline all the time to save memory. }
  if not FVisible then Exit;
  TmpVect := TVect2D.Create(FPointsInCurve);
  try
    GetCurvePoints(TmpVect);
    TmpFlag := FShowCtrlPts;
    { The draw polyline metod draw also the control points. }
    FShowCtrlPts := False;
    TmpTransf := MultiplyTransf2D(FModelTransform, VT);
    DrawPolyline(Cnv, TmpVect, TmpTransf);
    FShowCtrlPts := TmpFlag;
    if FShowCtrlPts then
     for Cont := 0 to FPoints.Count - 1 do
      begin
        TmpPt := WorldToView2D(FPoints[Cont], TmpTransf);
        DrawPlaceHolder(Cnv, Trunc(TmpPt.X), Trunc(TmpPt.Y), FCtrlPtsWidth);
      end;
  finally
    TmpVect.Free;
  end;
end;

function TBSpline2D.OnMe(Pt: TPoint2D; Aperture: TRealType;
                     var Distance: TRealType): Integer;
var
  TmpVect: TVect2D;
begin
  Result := -2;
  if not FEnabled then Exit;
  Result := PointOnPolyLine2D(Pt, FPoints, Distance, Aperture, FModelTransform);
  if Result < 0 then
   begin
     { Consider all segment in the spline only if the mouse is not on the
       polyline. }
     TmpVect := TVect2D.Create(FPointsInCurve);
     try
       GetCurvePoints(TmpVect);
       Result := PointOnPolyLine2D(Pt, TmpVect, Distance, Aperture, FModelTransform);
     finally
      TmpVect.Free;
     end;
     if Result > -2 then Result := -1;
   end;
end;

begin
  RegisterClass2D(TBSpline2D);
end.
